@@ -922,8 +922,17 @@ void hostap_setup_dev(struct net_device *dev, local_info_t *local,
 
 	/* kernel callbacks */
 	dev->get_stats = hostap_get_stats;
+#if WIRELESS_EXT > 16
+	if (iface) {
+		/* Currently, we point to the proper spy_data only on
+		 * the main_dev. This could be fixed. Jean II */
+		iface->wireless_data.spy_data = &iface->spy_data;
+		dev->wireless_data = &iface->wireless_data;
+	}
+#else /* WIRELESS_EXT > 16 */
 	if (main_dev || (iface && iface->type == HOSTAP_INTERFACE_MAIN))
 		dev->get_wireless_stats =  hostap_get_wireless_stats;
+#endif /* WIRELESS_EXT > 16 */
 	dev->wireless_handlers =
 		(struct iw_handler_def *) &hostap_iw_handler_def;
 	dev->do_ioctl = hostap_ioctl;
