@@ -32,6 +32,7 @@
 #ifdef HOSTAP_USE_CRYPTO_API
 #include <linux/crypto.h>
 #include <asm/scatterlist.h>
+#include <linux/crc32.h>
 #endif /* HOSTAP_USE_CRYPTO_API */
 #else /* CONFIG_CRYPTO */
 #undef HOSTAP_USE_CRYPTO_API
@@ -66,10 +67,15 @@ struct hostap_tkip_data {
 	int key_idx;
 
 #ifdef HOSTAP_USE_CRYPTO_API
+	struct crypto_tfm *tfm_arc4;
 	struct crypto_tfm *tfm_michael;
-#endif
+#endif /* HOSTAP_USE_CRYPTO_API */
+
+	/* scratch buffers for virt_to_page() (crypto API) */
+	u8 rx_hdr[16], tx_hdr[16];
 };
 
+#ifndef HOSTAP_USE_CRYPTO_API
 static const __u32 crc32_table[256] = {
 	0x00000000L, 0x77073096L, 0xee0e612cL, 0x990951baL, 0x076dc419L,
 	0x706af48fL, 0xe963a535L, 0x9e6495a3L, 0x0edb8832L, 0x79dcb8a4L,
@@ -124,6 +130,7 @@ static const __u32 crc32_table[256] = {
 	0x5d681b02L, 0x2a6f2b94L, 0xb40bbe37L, 0xc30c8ea1L, 0x5a05df1bL,
 	0x2d02ef8dL
 };
+#endif /* HOSTAP_USE_CRYPTO_API */
 
 
 static void * hostap_tkip_init(int key_idx)
@@ -144,18 +151,33 @@ static void * hostap_tkip_init(int key_idx)
 	priv->key_idx = key_idx;
 
 #ifdef HOSTAP_USE_CRYPTO_API
+	priv->tfm_arc4 = crypto_alloc_tfm("arc4", 0);
+	if (priv->tfm_arc4 == NULL) {
+		printk(KERN_DEBUG "hostap_crypt_tkip: could not allocate "
+		       "crypto API arc4\n");
+		goto fail;
+	}
+
 	priv->tfm_michael = crypto_alloc_tfm("michael_mic", 0);
 	if (priv->tfm_michael == NULL) {
 		printk(KERN_DEBUG "hostap_crypt_tkip: could not allocate "
 		       "crypto API michael_mic\n");
 		goto fail;
 	}
-#endif
+#endif /* HOSTAP_USE_CRYPTO_API */
 
 	return priv;
 
 fail:
-	kfree(priv);
+	if (priv) {
+#ifdef HOSTAP_USE_CRYPTO_API
+		if (priv->tfm_michael)
+			crypto_free_tfm(priv->tfm_michael);
+		if (priv->tfm_arc4)
+			crypto_free_tfm(priv->tfm_arc4);
+#endif /* HOSTAP_USE_CRYPTO_API */
+		kfree(priv);
+	}
 #ifdef NEW_MODULE_CODE
 	module_put(THIS_MODULE);
 #else
@@ -171,7 +193,9 @@ static void hostap_tkip_deinit(void *priv)
 	struct hostap_tkip_data *_priv = priv;
 	if (_priv && _priv->tfm_michael)
 		crypto_free_tfm(_priv->tfm_michael);
-#endif
+	if (_priv && _priv->tfm_arc4)
+		crypto_free_tfm(_priv->tfm_arc4);
+#endif /* HOSTAP_USE_CRYPTO_API */
 	kfree(priv);
 #ifdef NEW_MODULE_CODE
 	module_put(THIS_MODULE);
@@ -338,6 +362,7 @@ static void tkip_mixing_phase2(u8 *WEPSeed, const u8 *TK, const u16 *TTAK,
 }
 
 
+#ifndef HOSTAP_USE_CRYPTO_API
 static void hostap_wep_encrypt(u8 *key, u8 *buf, size_t buflen, u8 *icv)
 {
 	u32 i, j, k, crc;
@@ -380,6 +405,7 @@ static void hostap_wep_encrypt(u8 *key, u8 *buf, size_t buflen, u8 *icv)
 		*pos++ ^= S[(S[i] + S[j]) & 0xff];
 	}
 }
+#endif /* HOSTAP_USE_CRYPTO_API */
 
 
 static int hostap_tkip_encrypt(struct sk_buff *skb, int hdr_len, void *priv)
@@ -388,6 +414,10 @@ static int hostap_tkip_encrypt(struct sk_buff *skb, int hdr_len, void *priv)
 	int len;
 	u8 rc4key[16], *pos, *icv;
 	struct hostap_ieee80211_hdr *hdr;
+#ifdef HOSTAP_USE_CRYPTO_API
+	u32 crc;
+	struct scatterlist sg;
+#endif /* HOSTAP_USE_CRYPTO_API */
 
 	if (skb_headroom(skb) < 8 || skb_tailroom(skb) < 4 ||
 	    skb->len < hdr_len)
@@ -416,7 +446,21 @@ static int hostap_tkip_encrypt(struct sk_buff *skb, int hdr_len, void *priv)
 	*pos++ = (tkey->tx_iv32 >> 16) & 0xff;
 	*pos++ = (tkey->tx_iv32 >> 24) & 0xff;
 
+#ifdef HOSTAP_USE_CRYPTO_API
+	crc = ~crc32_le(~0, pos, len);
+	icv[0] = crc;
+	icv[1] = crc >> 8;
+	icv[2] = crc >> 16;
+	icv[3] = crc >> 24;
+
+	crypto_cipher_setkey(tkey->tfm_arc4, rc4key, 16);
+	sg.page = virt_to_page(pos);
+	sg.offset = offset_in_page(pos);
+	sg.length = len + 4;
+	crypto_cipher_encrypt(tkey->tfm_arc4, &sg, &sg, len + 4);
+#else /* HOSTAP_USE_CRYPTO_API */
 	hostap_wep_encrypt(rc4key, pos, len, icv);
+#endif /* HOSTAP_USE_CRYPTO_API */
 
 	tkey->tx_iv16++;
 	if (tkey->tx_iv16 == 0) {
@@ -428,6 +472,7 @@ static int hostap_tkip_encrypt(struct sk_buff *skb, int hdr_len, void *priv)
 }
 
 
+#ifndef HOSTAP_USE_CRYPTO_API
 static int hostap_wep_decrypt(u8 *key, u8 *buf, size_t plen)
 {
 	u32 i, j, k, crc;
@@ -475,6 +520,7 @@ static int hostap_wep_decrypt(u8 *key, u8 *buf, size_t plen)
 
 	return 0;
 }
+#endif /* HOSTAP_USE_CRYPTO_API */
 
 
 static int hostap_tkip_decrypt(struct sk_buff *skb, int hdr_len, void *priv)
@@ -485,6 +531,12 @@ static int hostap_tkip_decrypt(struct sk_buff *skb, int hdr_len, void *priv)
 	u32 iv32;
 	u16 iv16;
 	struct hostap_ieee80211_hdr *hdr;
+#ifdef HOSTAP_USE_CRYPTO_API
+	u8 icv[4];
+	u32 crc;
+	struct scatterlist sg;
+	int plen;
+#endif /* HOSTAP_USE_CRYPTO_API */
 
 	if (skb->len < hdr_len + 8 + 4)
 		return -1;
@@ -535,7 +587,24 @@ static int hostap_tkip_decrypt(struct sk_buff *skb, int hdr_len, void *priv)
 	}
 	tkip_mixing_phase2(rc4key, tkey->key, tkey->rx_ttak, iv16);
 
+#ifdef HOSTAP_USE_CRYPTO_API
+	plen = skb->len - hdr_len - 12;
+
+	crypto_cipher_setkey(tkey->tfm_arc4, rc4key, 16);
+	sg.page = virt_to_page(pos);
+	sg.offset = offset_in_page(pos);
+	sg.length = plen + 4;
+	crypto_cipher_decrypt(tkey->tfm_arc4, &sg, &sg, plen + 4);
+
+	crc = ~crc32_le(~0, pos, plen);
+	icv[0] = crc;
+	icv[1] = crc >> 8;
+	icv[2] = crc >> 16;
+	icv[3] = crc >> 24;
+	if (memcmp(icv, pos + plen, 4) != 0) {
+#else /* HOSTAP_USE_CRYPTO_API */
 	if (hostap_wep_decrypt(rc4key, pos, skb->len - hdr_len - 12)) {
+#endif /* HOSTAP_USE_CRYPTO_API */
 		if (iv32 != tkey->rx_iv32) {
 			/* Previously cached Phase1 result was already lost, so
 			 * it needs to be recalculated for the next packet. */
@@ -725,7 +794,6 @@ static int hostap_michael_mic_add(struct sk_buff *skb, int hdr_len, void *priv)
 {
 	struct hostap_tkip_data *tkey = priv;
 	u8 *pos;
-	u8 hdr[16];
 
 	if (skb_tailroom(skb) < 8 || skb->len < hdr_len) {
 		printk(KERN_DEBUG "Invalid packet for Michael MIC add "
@@ -734,9 +802,9 @@ static int hostap_michael_mic_add(struct sk_buff *skb, int hdr_len, void *priv)
 		return -1;
 	}
 
-	michael_mic_hdr(skb, hdr);
+	michael_mic_hdr(skb, tkey->tx_hdr);
 	pos = skb_put(skb, 8);
-	if (michael_mic(tkey, &tkey->key[16], hdr,
+	if (michael_mic(tkey, &tkey->key[16], tkey->tx_hdr,
 			skb->data + hdr_len, skb->len - 8 - hdr_len, pos))
 		return -1;
 
@@ -744,7 +812,28 @@ static int hostap_michael_mic_add(struct sk_buff *skb, int hdr_len, void *priv)
 }
 
 
-#if WIRELESS_EXT >= 15
+#if WIRELESS_EXT >= 18
+static void hostap_michael_mic_failure(struct net_device *dev,
+				       struct hostap_ieee80211_hdr *hdr,
+				       int keyidx)
+{
+	union iwreq_data wrqu;
+	struct iw_michaelmicfailure ev;
+
+	/* TODO: needed parameters: count, keyid, key type, TSC */
+	memset(&ev, 0, sizeof(ev));
+	ev.flags = keyidx & IW_MICFAILURE_KEY_ID;
+	if (hdr->addr1[0] & 0x01)
+		ev.flags |= IW_MICFAILURE_GROUP;
+	else
+		ev.flags |= IW_MICFAILURE_PAIRWISE;
+	ev.src_addr.sa_family = ARPHRD_ETHER;
+	memcpy(ev.src_addr.sa_data, hdr->addr2, ETH_ALEN);
+	memset(&wrqu, 0, sizeof(wrqu));
+	wrqu.data.length = sizeof(ev);
+	wireless_send_event(dev, IWEVMICHAELMICFAILURE, &wrqu, (char *) &ev);
+}
+#elif WIRELESS_EXT >= 15
 static void hostap_michael_mic_failure(struct net_device *dev,
 				       struct hostap_ieee80211_hdr *hdr,
 				       int keyidx)
@@ -752,7 +841,7 @@ static void hostap_michael_mic_failure(struct net_device *dev,
 	union iwreq_data wrqu;
 	char buf[128];
 
-	/* TODO: needed parameters: count, keyid, key type, src address, TSC */
+	/* TODO: needed parameters: count, keyid, key type, TSC */
 	sprintf(buf, "MLME-MICHAELMICFAILURE.indication(keyid=%d %scast addr="
 		MACSTR ")", keyidx, hdr->addr1[0] & 0x01 ? "broad" : "uni",
 		MAC2STR(hdr->addr2));
@@ -773,14 +862,13 @@ static int hostap_michael_mic_verify(struct sk_buff *skb, int keyidx,
 				     int hdr_len, void *priv)
 {
 	struct hostap_tkip_data *tkey = priv;
-	u8 hdr[16];
 	u8 mic[8];
 
 	if (!tkey->key_set)
 		return -1;
 
-	michael_mic_hdr(skb, hdr);
-	if (michael_mic(tkey, &tkey->key[24], hdr,
+	michael_mic_hdr(skb, tkey->rx_hdr);
+	if (michael_mic(tkey, &tkey->key[24], tkey->rx_hdr,
 			skb->data + hdr_len, skb->len - 8 - hdr_len, mic))
 		return -1;
 	if (memcmp(mic, skb->data + skb->len - 8, 8) != 0) {
@@ -813,14 +901,16 @@ static int hostap_tkip_set_key(void *key, int len, u8 *seq, void *priv)
 	int keyidx;
 #ifdef HOSTAP_USE_CRYPTO_API
 	struct crypto_tfm *tfm = tkey->tfm_michael;
-#endif
+	struct crypto_tfm *tfm2 = tkey->tfm_arc4;
+#endif /* HOSTAP_USE_CRYPTO_API */
 
 	keyidx = tkey->key_idx;
 	memset(tkey, 0, sizeof(*tkey));
 	tkey->key_idx = keyidx;
 #ifdef HOSTAP_USE_CRYPTO_API
 	tkey->tfm_michael = tfm;
-#endif
+	tkey->tfm_arc4 = tfm2;
+#endif /* HOSTAP_USE_CRYPTO_API */
 	if (len == TKIP_KEY_LEN) {
 		memcpy(tkey->key, key, TKIP_KEY_LEN);
 		tkey->key_set = 1;
